"""
This module contains models for Beatmap objects.
"""

from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime
from enum import Enum
from enum import unique
from functools import cached_property
from typing import Literal
from typing import Optional

from pydantic import Field
from pydantic import computed_field
from pydantic import model_validator

from .base import BaseModel
from .base import cast_int
from .common import CurrentUserAttributes
from .common import CursorModel
from .gamemode import Gamemode
from .user import User

__all__ = (
    "Beatmap",
    "BeatmapAvailability",
    "BeatmapCovers",
    "BeatmapDescription",
    "BeatmapDifficultyAttributes",
    "BeatmapFailtimes",
    "BeatmapGenre",
    "BeatmapHype",
    "BeatmapLanguage",
    "BeatmapNominations",
    "BeatmapPack",
    "BeatmapPackType",
    "BeatmapPackUserCompletion",
    "BeatmapPacksResponse",
    "BeatmapRankStatus",
    "BeatmapUserPlaycount",
    "Beatmapset",
    "BeatmapsetBundleFilterType",
    "BeatmapsetCategory",
    "BeatmapsetDiscussion",
    "BeatmapsetDiscussionPost",
    "BeatmapsetDiscussionPostResponse",
    "BeatmapsetDiscussionResponse",
    "BeatmapsetDiscussionVoteResponse",
    "BeatmapsetDiscussionVoteScoreType",
    "BeatmapsetDisscussionType",
    "BeatmapsetEvent",
    "BeatmapsetEventComment",
    "BeatmapsetEventType",
    "BeatmapsetGenre",
    "BeatmapsetLanguage",
    "BeatmapsetRequestStatus",
    "BeatmapsetSearchResponse",
    "BeatmapsetSortType",
    "BeatmapsetVoteEvent",
    "UserBeatmapType",
)

BeatmapsetSortType = Literal[
    "title_asc",
    "title_desc",
    "artist_asc",
    "artist_desc",
    "difficulty_asc",
    "difficulty_desc",
    "ranked_asc",
    "ranked_desc",
    "rating_asc",
    "rating_desc",
    "plays_asc",
    "plays_desc",
    "favourites_asc",
    "favourites_desc",
]

BeatmapsetCategory = Literal[
    "any",
    "leaderboard",
    "ranked",
    "qualified",
    "loved",
    "favourites",
    "pending",
    "wip",
    "graveyard",
    "mine",
]

BeatmapsetBundleFilterType = Literal[
    "any",
    "currently",
    "previously",
    "never",
]

BeatmapsetDisscussionType = Literal[
    "hype",
    "praise",
    "problem",
    "review",
    "suggestion",
    "mapper_note",
]


BeatmapsetEventType = Literal[
    "approve",
    "beatmap_owner_change",
    "discussion_delete",
    "discussion_post_delete",
    "discussion_post_restore",
    "discussion_restore",
    "discussion_lock",
    "discussion_unlock",
    "disqualify",
    "genre_edit",
    "issue_reopen",
    "issue_resolve",
    "kudosu_allow",
    "kudosu_deny",
    "kudosu_gain",
    "kudosu_lost",
    "kudosu_recalculate",
    "language_edit",
    "love",
    "nominate",
    "nomination_reset",
    "nomination_reset_received",
    "nsfw_toggle",
    "offset_edit",
    "qualify",
    "rank",
    "remove_from_loved",
]

BeatmapsetRequestStatus = Literal[
    "all",
    "ranked",
    "qualified",
    "disqualified",
    "never_ranked",
]

BeatmapsetDiscussionVoteScoreType = Literal[
    "1",
    "-1",
]

UserBeatmapType = Literal["favourite", "graveyard", "loved", "ranked", "pending"]

BEATMAP_RANK_STATUS_NAMES = {
    -2: "graveyard",
    -1: "wip",
    0: "pending",
    1: "ranked",
    2: "approved",
    3: "qualified",
    4: "loved",
}


@unique
class BeatmapsetGenre(Enum):
    UNSPECIFIED = 1
    VIDEO_GAME = 2
    ANIME = 3
    ROCK = 4
    POP = 5
    OTHER = 6
    NOVELTY = 7
    HIP_HOP = 9
    ELECTRONIC = 10
    METAL = 11
    CLASSICAL = 12
    FOLK = 13
    JAZZ = 14

    @classmethod
    def _missing_(cls, _: object) -> BeatmapsetGenre:
        return BeatmapsetGenre.UNSPECIFIED


@unique
class BeatmapsetLanguage(Enum):
    UNSPECIFIED = 1
    ENGLISH = 2
    JAPANESE = 3
    CHINESE = 4
    INSTRUMENTAL = 5
    KOREAN = 6
    FRENCH = 7
    GERMAN = 8
    SWEDISH = 9
    SPANISH = 10
    ITALIAN = 11
    RUSSIAN = 12
    POLISH = 13
    OTHER = 14

    @classmethod
    def _missing_(cls, _: object) -> BeatmapsetLanguage:
        return BeatmapsetLanguage.UNSPECIFIED


@unique
class BeatmapRankStatus(Enum):
    GRAVEYARD = -2
    WIP = -1
    PENDING = 0
    RANKED = 1
    APPROVED = 2
    QUALIFIED = 3
    LOVED = 4

    @property
    def id(self) -> int:
        return self.value

    @property
    def name_api(self) -> str:
        return BEATMAP_RANK_STATUS_NAMES[self.id]

    def __str__(self) -> str:
        return self.name_api

    @classmethod
    def _missing_(cls, query: object) -> BeatmapRankStatus:
        if isinstance(query, int):
            for status in list(BeatmapRankStatus):
                if status.id == query:
                    return status
        elif isinstance(query, str):
            for status in list(BeatmapRankStatus):
                if status.name_api == query.lower():
                    return status
        raise ValueError(f"BeatmapRankStatus {query} does not exist.")


@unique
class BeatmapPackType(Enum):
    STANDARD = ("S", "Standard")
    FEATURED = ("F", "Featured Artist")
    TOURNAMENT = ("P", "Tournament")
    LOVED = ("L", "Project Loved")
    CHART = ("R", "Spotlights")
    THEME = ("T", "Theme")
    ARTIST = ("A", "Artist/Album")

    def __init__(self, tag: str, description: str) -> None:
        self.tag = tag
        self.description = description

    @classmethod
    def from_tag(cls, tag: str) -> BeatmapPackType:
        beatmap_pack_type = next((x for x in cls if x.tag == tag), None)
        if beatmap_pack_type is None:
            raise ValueError(f"BeatmapPackType {tag} does not exist.")
        return beatmap_pack_type

    def __str__(self) -> str:
        return self.name.lower()

    @classmethod
    def _missing_(cls, query: object) -> BeatmapPackType:
        if isinstance(query, cls):
            return query
        if isinstance(query, str):
            query = query.upper()
            try:
                return BeatmapPackType[query]
            except KeyError:
                return cls.from_tag(query)

        raise ValueError(f"BeatmapPackType {query} does not exist.")


class BeatmapDescription(BaseModel):
    bbcode: Optional[str] = None
    description: Optional[str] = None


class BeatmapGenre(BaseModel):
    name: str
    id: Optional[int] = None


class BeatmapLanguage(BaseModel):
    name: str
    id: Optional[int] = None


class BeatmapAvailability(BaseModel):
    more_information: Optional[str] = None
    download_disabled: Optional[bool] = None

    @classmethod
    def _from_api_v1(cls, data: Mapping[str, object]) -> BeatmapAvailability:
        return cls.model_validate({"download_disabled": data["download_unavailable"]})


class BeatmapNominations(BaseModel):
    current: Optional[int] = None
    required: Optional[int] = None


class BeatmapNomination(BaseModel):
    beatmapset_id: int
    reset: bool
    user_id: int
    rulesets: Optional[list[Gamemode]] = None


class BeatmapCovers(BaseModel):
    cover: str
    card: str
    list: str
    slimcover: str
    cover_2_x: Optional[str] = Field(default=None, alias="cover@2x")
    card_2_x: Optional[str] = Field(default=None, alias="card@2x")
    list_2_x: Optional[str] = Field(default=None, alias="list@2x")
    slimcover_2_x: Optional[str] = Field(default=None, alias="slimcover@2x")

    @classmethod
    def from_beatmapset_id(cls, beatmapset_id: int) -> BeatmapCovers:
        base_url = "https://assets.ppy.sh/beatmaps/"
        return cls.model_validate(
            {
                "cover": f"{base_url}{beatmapset_id}/covers/cover.jpg",
                "card": f"{base_url}{beatmapset_id}/covers/card.jpg",
                "list": f"{base_url}{beatmapset_id}/covers/list.jpg",
                "slimcover": f"{base_url}{beatmapset_id}/covers/slimcover.jpg",
                "cover_2_x": f"{base_url}{beatmapset_id}/covers/cover@2x.jpg",
                "card_2_x": f"{base_url}{beatmapset_id}/covers/card@2x.jpg",
                "list_2_x": f"{base_url}{beatmapset_id}/covers/list@2x.jpg",
                "slimcover_2_x": f"{base_url}{beatmapset_id}/covers/slimcover@2x.jpg",
            },
        )

    @classmethod
    def _from_api_v1(cls, data: Mapping[str, object]) -> BeatmapCovers:
        return cls.from_beatmapset_id(cast_int(data["beatmapset_id"]))


class BeatmapHype(BaseModel):
    current: int
    required: int


class BeatmapFailtimes(BaseModel):
    exit: Optional[list[int]] = None
    fail: Optional[list[int]] = None


class BeatmapDifficultyAttributes(BaseModel):
    max_combo: int
    star_rating: float
    # osu standard
    aim_difficulty: Optional[float] = None
    approach_rate: Optional[float] = None  # osu catch + standard
    flashlight_difficulty: Optional[float] = None
    overall_difficulty: Optional[float] = None
    slider_factor: Optional[float] = None
    speed_difficulty: Optional[float] = None
    speed_note_count: Optional[float] = None
    # osu taiko
    stamina_difficulty: Optional[float] = None
    rhythm_difficulty: Optional[float] = None
    colour_difficulty: Optional[float] = None
    # osu mania
    great_hit_window: Optional[float] = None
    score_multiplier: Optional[float] = None


class Beatmap(BaseModel):
    id: int
    url: str
    mode: Gamemode
    beatmapset_id: int
    difficulty_rating: float
    status: BeatmapRankStatus
    total_length: int
    user_id: int
    version: str
    accuracy: Optional[float] = None
    ar: Optional[float] = None
    cs: Optional[float] = None
    bpm: Optional[float] = None
    convert: Optional[bool] = None
    count_circles: Optional[int] = None
    count_sliders: Optional[int] = None
    count_spinners: Optional[int] = None
    deleted_at: Optional[datetime] = None
    drain: Optional[float] = None
    hit_length: Optional[int] = None
    is_scoreable: Optional[bool] = None
    last_updated: Optional[datetime] = None
    passcount: Optional[int] = None
    play_count: Optional[int] = Field(default=None, alias="playcount")
    checksum: Optional[str] = None
    max_combo: Optional[int] = None
    beatmapset: Optional[Beatmapset] = None
    failtimes: Optional[BeatmapFailtimes] = None

    @property
    def discussion_url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.beatmapset_id}/discussion/{self.id}/general"

    @computed_field  # type: ignore
    @cached_property
    def count_objects(self) -> Optional[int]:
        """Total count of the objects.

        :return: Sum of counts of all objects. None if no object count information.
        :rtype: Optional[int]
        """
        if (
            self.count_circles is None
            or self.count_spinners is None
            or self.count_sliders is None
        ):
            return None
        return self.count_spinners + self.count_circles + self.count_sliders

    @model_validator(mode="before")
    @classmethod
    def _set_url(cls, values: dict[str, object]) -> dict[str, object]:
        if values.get("url") is None:
            id = values["id"]
            beatmapset_id = values["beatmapset_id"]
            mode = Gamemode(values["mode"])
            values["url"] = (
                f"https://osu.ppy.sh/beatmapsets/{beatmapset_id}#{mode}/{id}"
            )
        return values

    @classmethod
    def _from_api_v1(cls, data: Mapping[str, object]) -> Beatmap:
        return cls.model_validate(
            {
                "beatmapset_id": data["beatmapset_id"],
                "difficulty_rating": data["difficultyrating"],
                "id": data["beatmap_id"],
                "mode": cast_int(data["mode"]),
                "status": cast_int(data["approved"]),
                "total_length": data["total_length"],
                "hit_length": data["total_length"],
                "user_id": data["creator_id"],
                "version": data["version"],
                "accuracy": data["diff_overall"],
                "cs": data["diff_size"],
                "ar": data["diff_approach"],
                "drain": data["diff_drain"],
                "last_updated": data["last_update"],
                "bpm": data["bpm"],
                "checksum": data["file_md5"],
                "playcount": data["playcount"],
                "passcount": data["passcount"],
                "count_circles": data["count_normal"],
                "count_sliders": data["count_slider"],
                "count_spinners": data["count_spinner"],
                "max_combo": data["max_combo"],
            },
        )


class Beatmapset(BaseModel):
    id: int
    artist: str
    artist_unicode: str
    covers: BeatmapCovers
    creator: str
    favourite_count: int
    play_count: int = Field(alias="playcount")
    preview_url: str
    source: str
    status: BeatmapRankStatus
    title: str
    title_unicode: str
    user_id: int
    video: bool
    nsfw: Optional[bool] = None
    hype: Optional[BeatmapHype] = None
    availability: Optional[BeatmapAvailability] = None
    bpm: Optional[float] = None
    can_be_hyped: Optional[bool] = None
    discussion_enabled: Optional[bool] = None
    discussion_locked: Optional[bool] = None
    is_scoreable: Optional[bool] = None
    last_updated: Optional[datetime] = None
    legacy_thread_url: Optional[str] = None
    nominations: Optional[BeatmapNominations] = None
    current_nominations: Optional[list[BeatmapNomination]] = None
    ranked_date: Optional[datetime] = None
    storyboard: Optional[bool] = None
    submitted_date: Optional[datetime] = None
    tags: Optional[str] = None
    pack_tags: Optional[list[str]] = None
    track_id: Optional[int] = None
    user: Optional[User] = None
    related_users: Optional[list[User]] = None
    current_user_attributes: Optional[CurrentUserAttributes] = None
    description: Optional[BeatmapDescription] = None
    genre: Optional[BeatmapGenre] = None
    language: Optional[BeatmapLanguage] = None
    ratings: Optional[list[int]] = None
    recent_favourites: Optional[list[User]] = None
    discussions: Optional[list[BeatmapsetDiscussion]] = None
    events: Optional[list[BeatmapsetEvent]] = None
    has_favourited: Optional[bool] = None
    beatmaps: Optional[list[Beatmap]] = None
    converts: Optional[list[Beatmap]] = None

    @computed_field  # type: ignore
    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.id}"

    @computed_field  # type: ignore
    @property
    def discussion_url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.id}/discussion"

    @classmethod
    def _from_api_v1(cls, data: Mapping[str, object]) -> Beatmapset:
        return cls.model_validate(
            {
                "id": data["beatmapset_id"],
                "artist": data["artist"],
                "artist_unicode": data["artist"],
                "covers": BeatmapCovers._from_api_v1(data),
                "favourite_count": data["favourite_count"],
                "creator": data["creator"],
                "play_count": data["playcount"],
                "preview_url": f"//b.ppy.sh/preview/{data['beatmapset_id']}.mp3",
                "source": data["source"],
                "status": cast_int(data["approved"]),
                "title": data["title"],
                "title_unicode": data["title"],
                "user_id": data["creator_id"],
                "video": data["video"],
                "submitted_date": data["submit_date"],
                "ranked_date": data["approved_date"],
                "last_updated": data["last_update"],
                "tags": data["tags"],
                "storyboard": data["storyboard"],
                "availabiliy": BeatmapAvailability._from_api_v1(data),
                "beatmaps": [Beatmap._from_api_v1(data)],
            },
        )


class BeatmapsetSearchResponse(CursorModel):
    beatmapsets: list[Beatmapset]


class BeatmapUserPlaycount(BaseModel):
    count: int
    beatmap_id: int
    beatmap: Optional[Beatmap] = None
    beatmapset: Optional[Beatmapset] = None


class BeatmapsetDiscussionPost(BaseModel):
    id: int
    user_id: int
    system: bool
    message: str
    created_at: datetime
    beatmap_discussion_id: Optional[int] = None
    last_editor_id: Optional[int] = None
    deleted_by_id: Optional[int] = None
    updated_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None


class BeatmapsetDiscussion(BaseModel):
    id: int
    beatmapset_id: int
    user_id: int
    message_type: BeatmapsetDisscussionType
    resolved: bool
    can_be_resolved: bool
    can_grant_kudosu: bool
    created_at: datetime
    beatmap_id: Optional[int] = None
    deleted_by_id: Optional[int] = None
    parent_id: Optional[int] = None
    timestamp: Optional[int] = None
    updated_at: Optional[datetime] = None
    deleted_at: Optional[datetime] = None
    last_post_at: Optional[datetime] = None
    kudosu_denied: Optional[bool] = None
    starting_post: Optional[BeatmapsetDiscussionPost] = None


class BeatmapsetVoteEvent(BaseModel):
    score: int
    user_id: int
    id: Optional[int] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    beatmapset_discussion_id: Optional[int] = None


class BeatmapsetEventComment(BaseModel):
    beatmap_discussion_id: Optional[int] = None
    beatmap_discussion_post_id: Optional[int] = None
    new_vote: Optional[BeatmapsetVoteEvent] = None
    votes: Optional[list[BeatmapsetVoteEvent]] = None
    mode: Optional[Gamemode] = None
    reason: Optional[str] = None
    source_user_id: Optional[int] = None
    source_user_username: Optional[str] = None
    nominator_ids: Optional[list[int]] = None
    new: Optional[str] = None
    old: Optional[str] = None
    new_user_id: Optional[int] = None
    new_user_username: Optional[str] = None


class BeatmapsetEvent(BaseModel):
    id: int
    type: BeatmapsetEventType
    r"""Information on types: https://github.com/ppy/osu-web/blob/master/resources/js/interfaces/beatmapset-event-json.ts"""
    created_at: datetime
    user_id: Optional[int] = None
    beatmapset: Optional[Beatmapset] = None
    discussion: Optional[BeatmapsetDiscussion] = None
    comment: Optional[dict] = None


class BeatmapPackUserCompletion(BaseModel):
    beatmapset_ids: list[int]
    completed: bool


class BeatmapPack(BaseModel):
    author: str
    date: datetime
    name: str
    no_diff_reduction: bool
    tag: str
    url: str
    ruleset_id: Optional[int] = None
    beatmapsets: Optional[list[Beatmapset]] = None
    user_completion_data: Optional[BeatmapPackUserCompletion] = None

    @property
    def mode(self) -> Optional[Gamemode]:
        if self.ruleset_id is None:
            return None
        return Gamemode(self.ruleset_id)

    @property
    def pack_type(self) -> BeatmapPackType:
        return BeatmapPackType.from_tag(self.tag[0])

    @property
    def id(self) -> int:
        return int(self.tag[1:])


class BeatmapPacksResponse(CursorModel):
    beatmap_packs: list[BeatmapPack]


class BeatmapsetDiscussionResponse(CursorModel):
    beatmaps: list[Beatmap]
    discussions: list[BeatmapsetDiscussion]
    included_discussions: list[BeatmapsetDiscussion]
    users: list[User]
    max_blocks: int

    @model_validator(mode="before")
    @classmethod
    def _set_max_blocks(cls, values: dict[str, object]) -> dict[str, object]:
        if isinstance(values["reviews_config"], Mapping):
            values["max_blocks"] = values["reviews_config"]["max_blocks"]

        return values


class BeatmapsetDiscussionPostResponse(CursorModel):
    beatmapsets: list[Beatmapset]
    posts: list[BeatmapsetDiscussionPost]
    users: list[User]


class BeatmapsetDiscussionVoteResponse(CursorModel):
    votes: list[BeatmapsetVoteEvent]
    discussions: list[BeatmapsetDiscussion]
    users: list[User]



"""
This module contains miscellaneous helper functions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping
    from collections.abc import MutableMapping
    from typing import Any
    from typing import Callable
    from typing import Optional
    from typing import TypeVar

    T = TypeVar("T")

__all__ = (
    "add_param",
    "add_range",
    "append_param",
    "from_list",
)


def from_list(f: Callable[[Any], T], x: list) -> list[T]:
    r"""Applies a function to all elements in a list.

    :param f: Function to apply on list elements
    :type f: Callable[[Any], T]
    :param x: List of objects
    :type x: list[object]
    :raises TypeError: If x is not a list
    :return: New list
    :rtype: list[T]
    """
    if not isinstance(x, list):
        raise TypeError("Wrong type received. Expected list.")
    return [f(y) for y in x]


def append_param(
    value: object,
    l: list,
    append: bool = True,
) -> None:
    r"""Appends a value to a list if it is not None and append is True.

    :param value: Value to append
    :type value: object
    :param l: List to append to
    :type l: list[object]
    :param append: Whether to append or not, defaults to True
    :type append: bool, optional, defaults to True
    """
    if value is None:
        return
    if append:
        l.append(value)


def add_param(
    params: MutableMapping[str, Any],
    kwargs: Mapping[str, object],
    key: str,
    param_name: Optional[str] = None,
    converter: Optional[Callable[[Any], T]] = None,
) -> bool:
    r"""Adds a parameter to a dictionary if it exists in kwargs.

    :param params: Dictionary to add parameter to
    :type params: Mapping[str, Any]
    :param kwargs: Dictionary to get parameter from
    :type kwargs: Mapping[str, Any]
    :param key: Key to get parameter from
    :type key: str
    :param param_name: Name of parameter to add to dictionary, defaults to None
    :type param_name: Optional[str]
    :param converter: Function to convert parameter, defaults to None
    :type converter: Optional[Callable[[Any], T]]
    :return: True if parameter was added, False otherwise
    :rtype: bool
    """
    if key not in kwargs:
        return False

    value = kwargs[key]
    if converter:
        value = converter(value)

    params[param_name or key] = value
    return True


def add_range(
    params: MutableMapping[str, Any],
    kwargs: Mapping[str, object],
    key: str,
    param_name: Optional[str] = None,
) -> None:
    r"""Adds a range parameter to a dictionary if it exists in kwargs.

    :param params: Dictionary to add parameter to
    :type params: Mapping[str, Any]
    :param kwargs: Dictionary to get parameter from
    :type kwargs: Mapping[str, Any]
    :param key: Key to get parameter from
    :type key: str
    :param param_name: Name of parameter to add to dictionary, defaults to None
    :type param_name: Optional[str]
    """
    if key not in kwargs:
        return

    value = kwargs[key]
    if not isinstance(value, (tuple, list)) or len(value) != 2:
        raise ValueError("Range parameter must be a tuple of length 2.")

    params[f"{param_name}[gte]"] = value[0]
    params[f"{param_name}[lte]"] = value[1]



"""
This module contains functions to parse replay files.
"""

from __future__ import annotations

from typing import Any
from typing import BinaryIO

from ..models.files.replay import ReplayEvent
from ..models.files.replay import ReplayFile
from ..models.files.replay import ReplayKey
from ..models.files.replay import ReplayLifebarEvent
from ..models.lazer import LazerReplayData
from ..models.mods import Mod
from .binary import pack_byte
from .binary import pack_float64
from .binary import pack_int
from .binary import pack_long
from .binary import pack_replay_data
from .binary import pack_short
from .binary import pack_string
from .binary import pack_timestamp
from .binary import unpack_byte
from .binary import unpack_float64
from .binary import unpack_int
from .binary import unpack_long
from .binary import unpack_replay_data
from .binary import unpack_short
from .binary import unpack_string
from .binary import unpack_timestamp

__all__ = (
    "parse_file",
    "parse_path",
    "write_path",
    "write_replay",
)


def _parse_replay_data(data: str) -> list[ReplayEvent]:
    """Parse replay event data and return a list of replay events."""
    events: list[ReplayEvent] = []
    for event in data.split(","):
        if event == "":
            continue
        event_data: list[str] = event.split("|")
        time: int = int(event_data[0])
        x: float = float(event_data[1])
        y: float = float(event_data[2])
        keys: ReplayKey = ReplayKey(int(event_data[3]))
        events.append(ReplayEvent(time=time, keys=keys, x=x, y=y))
    return events


def _parse_life_graph_data(data: str) -> list[ReplayLifebarEvent]:
    """Parse life bar data and return a list of life bar events."""
    events: list[ReplayLifebarEvent] = []
    for event in data.split(","):
        if event == "":
            continue
        event_data: list[str] = event.split("|")
        time: int = int(event_data[0])
        hp: float = float(event_data[1])
        events.append(ReplayLifebarEvent(time=time, hp=hp))
    return events


def parse_file(file: BinaryIO) -> ReplayFile:
    """Parse a replay file and return a dictionary with the replay data.

    :param file: The replay file.
    :type file: BinaryIO
    :return: The replay data.
    :rtype: Replay
    """
    replay: dict[str, Any] = {}
    statistics: dict[str, int] = {}
    replay["mode"] = unpack_byte(file)
    replay["version"] = unpack_int(file)
    replay["map_md5"] = unpack_string(file)
    replay["player_name"] = unpack_string(file)
    replay["replay_md5"] = unpack_string(file)
    statistics["count_300"] = unpack_short(file)
    statistics["count_100"] = unpack_short(file)
    statistics["count_50"] = unpack_short(file)
    statistics["count_geki"] = unpack_short(file)
    statistics["count_katu"] = unpack_short(file)
    statistics["count_miss"] = unpack_short(file)
    replay["statistics"] = statistics
    replay["score"] = unpack_int(file)
    replay["max_combo"] = unpack_short(file)
    replay["perfect_combo"] = unpack_byte(file)
    replay["mods"] = unpack_int(file)
    lifebar_data_str = unpack_string(file)
    replay["lifebar_data"] = _parse_life_graph_data(lifebar_data_str)
    replay["played_at"] = unpack_timestamp(file)
    replay_data_str = unpack_replay_data(file)
    replay["replay_data"] = _parse_replay_data(replay_data_str)
    if replay["version"] >= 20140721:
        replay["online_id"] = unpack_long(file)
    elif replay["version"] >= 20121008:
        replay["online_id"] = unpack_int(file)
    if Mod.Target & replay["mods"]:
        replay["mod_extras"] = unpack_float64(file)
    if replay["version"] >= 30000001:
        lazer_replay_data_str = unpack_replay_data(file)
        replay["lazer_replay_data"] = LazerReplayData.model_validate_json(
            lazer_replay_data_str,
        )
    return ReplayFile(**replay)


def parse_path(path: str) -> ReplayFile:
    """Parse a replay file and return a dictionary with the replay data.

    :param path: The path to the replay file.
    :type path: str
    :return: The replay data.
    :rtype: Replay
    """
    with open(path, "rb") as file:
        return parse_file(file)


def write_replay(file: BinaryIO, replay: ReplayFile) -> None:
    """Write a replay to a file.

    :param file: The file to write to.
    :type file: BinaryIO
    :param replay: The replay to write.
    :type replay: Replay
    """
    pack_byte(file, int(replay.mode))
    pack_int(file, replay.version)
    pack_string(file, replay.map_md5)
    pack_string(file, replay.player_name)
    pack_string(file, replay.replay_md5)
    pack_short(file, replay.statistics.count_300)
    pack_short(file, replay.statistics.count_100)
    pack_short(file, replay.statistics.count_50)
    pack_short(file, replay.statistics.count_geki)
    pack_short(file, replay.statistics.count_katu)
    pack_short(file, replay.statistics.count_miss)
    pack_int(file, replay.score)
    pack_short(file, replay.max_combo)
    pack_byte(file, replay.perfect_combo)
    pack_int(file, int(replay.mods))
    pack_string(
        file,
        ",".join(
            [f"{event.time}|{event.hp}" for event in replay.lifebar_data],
        ),
    )
    pack_timestamp(file, replay.played_at)
    pack_replay_data(
        file,
        ",".join(
            [
                f"{event.time}|{event.x}|{event.y}|{event.keys}"
                for event in replay.replay_data
            ],
        ),
    )
    if replay.version >= 2014_07_21:
        pack_long(file, replay.online_id)
    elif replay.version >= 2012_10_08:
        pack_int(file, replay.online_id)
    if replay.mod_extras is not None:
        pack_float64(file, replay.mod_extras)
    if replay.lazer_replay_data is not None:
        pack_replay_data(
            file,
            replay.lazer_replay_data.model_dump_json(
                exclude_unset=True,
                exclude_none=True,
            ),
        )


def write_path(path: str, replay: ReplayFile) -> None:
    """Write a replay to a file.

    :param path: The path to the file to write to.
    :type path: str
    :param replay: The replay to write.
    :type replay: Replay
    """
    with open(path, "wb") as file:
        write_replay(file, replay)



"""
This module contains functions for reading and writing binary data.
"""

from __future__ import annotations

import lzma
import struct
from datetime import datetime
from datetime import timezone
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:
    from typing import BinaryIO

_lzma_format = lzma.FORMAT_ALONE

__all__ = (
    "pack",
    "pack_byte",
    "pack_float16",
    "pack_float32",
    "pack_float64",
    "pack_int",
    "pack_long",
    "pack_replay_data",
    "pack_short",
    "pack_string",
    "pack_timestamp",
    "pack_uleb128",
    "unpack",
    "unpack_byte",
    "unpack_float16",
    "unpack_float32",
    "unpack_float64",
    "unpack_int",
    "unpack_long",
    "unpack_replay_data",
    "unpack_short",
    "unpack_string",
    "unpack_timestamp",
    "unpack_uleb128",
)


def unpack(file: BinaryIO, fmt: str) -> int:
    r"""Unpack a value from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :param fmt: The format to unpack.
    :type fmt: str
    :return: The unpacked value.
    :rtype: int
    """
    return struct.unpack(fmt, file.read(struct.calcsize(fmt)))[0]


def unpack_byte(file: BinaryIO) -> int:
    r"""Unpack a byte from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked byte.
    :rtype: int
    """
    return unpack(file, "<b")


def unpack_short(file: BinaryIO) -> int:
    r"""Unpack a short from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked short.
    :rtype: int
    """
    return unpack(file, "<h")


def unpack_int(file: BinaryIO) -> int:
    r"""Unpack an integer from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked integer.
    :rtype: int
    """
    return unpack(file, "<i")


def unpack_long(file: BinaryIO) -> int:
    r"""Unpack a long from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked long.
    :rtype: int
    """
    return unpack(file, "q")


def unpack_float16(file: BinaryIO) -> float:
    r"""Unpack a float16 from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked float16.
    :rtype: float
    """
    return unpack(file, "<e")


def unpack_float32(file: BinaryIO) -> float:
    r"""Unpack a float32 from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked float32.
    :rtype: float
    """
    return unpack(file, "<f")


def unpack_float64(file: BinaryIO) -> float:
    r"""Unpack a float64 from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked float64.
    :rtype: float
    """
    return unpack(file, "<d")


def unpack_timestamp(file: BinaryIO) -> datetime:
    r"""Unpack a timestamp from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked timestamp.
    :rtype: datetime
    """
    seconds = unpack_long(file) // 10000000 - 62135596800
    return datetime.fromtimestamp(seconds, tz=timezone.utc)


def unpack_uleb128(file: BinaryIO) -> int:
    r"""Unpack a ULEB128 from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked ULEB128.
    :rtype: int
    """
    result = 0
    shift = 0
    while True:
        byte = unpack_byte(file)
        result |= (byte & 0x7F) << shift
        if not byte & 0x80:
            break
        shift += 7
    return result


def unpack_string(file: BinaryIO) -> str:
    r"""Unpack a string from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The unpacked string.
    :rtype: str
    """
    fb = file.read(1)
    if fb == b"\x00":
        return ""
    length = unpack_uleb128(file)
    return file.read(length).decode("utf-8")


def unpack_replay_data(file: BinaryIO) -> str:
    r"""Unpack the replay data from a file.

    :param file: The file to unpack from.
    :type file: BinaryIO
    :return: The replay event data.
    :rtype: str
    """
    length = unpack_int(file)
    if length == 0:
        return ""
    data = file.read(length)
    data = lzma.decompress(data)
    return data.decode("ascii")


def pack(file: BinaryIO, fmt: str, value: object) -> None:
    r"""Pack a value into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param fmt: The format to pack with.
    :type fmt: str
    :param value: The value to pack.
    :type value: object
    """
    file.write(struct.pack(fmt, value))


def pack_byte(file: BinaryIO, value: int) -> None:
    r"""Pack a byte into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: int
    """
    pack(file, "<b", value)


def pack_short(file: BinaryIO, value: int) -> None:
    r"""Pack a short into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: int
    """
    pack(file, "<h", value)


def pack_int(file: BinaryIO, value: int) -> None:
    r"""Pack an integer into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: int
    """
    pack(file, "<i", value)


def pack_long(file: BinaryIO, value: int) -> None:
    r"""Pack a long into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: int
    """
    pack(file, "<q", value)


def pack_float16(file: BinaryIO, value: float) -> None:
    r"""Pack a float16 into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: float
    """
    pack(file, "<e", value)


def pack_float32(file: BinaryIO, value: float) -> None:
    r"""Pack a float32 into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: float
    """
    pack(file, "<f", value)


def pack_float64(file: BinaryIO, value: float) -> None:
    r"""Pack a float64 into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: float
    """
    pack(file, "<d", value)


def pack_timestamp(file: BinaryIO, value: datetime) -> None:
    r"""Pack a timestamp into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: datetime
    """
    seconds = (value.timestamp() + 62135596800) * 10000000
    pack_long(file, int(seconds))


def pack_uleb128(file: BinaryIO, value: int) -> None:
    r"""Pack a ULEB128 into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: int
    """
    while True:
        byte = value & 0x7F
        value >>= 7
        if value:
            byte |= 0x80
        pack_byte(file, byte)
        if not value:
            break


def pack_string(file: BinaryIO, value: Union[bytes, str]) -> None:
    r"""Pack a string into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param value: The value to pack.
    :type value: Union[bytes, str]
    """
    pack_byte(file, 11)
    if not value:
        file.write(b"\x00")
        return
    pack_uleb128(file, len(value))
    file.write(value.encode("utf-8") if isinstance(value, str) else value)


def pack_replay_data(file: BinaryIO, data: str) -> None:
    r"""Pack the replay data into a file.

    :param file: The file to pack into.
    :type file: BinaryIO
    :param data: The data to pack.
    :type data: str
    """
    encoded_data = data.encode("ascii")
    compressed = lzma.compress(encoded_data, format=_lzma_format)
    pack_int(file, len(compressed))
    file.write(compressed)



# isort: dont-add-imports

from datetime import date
from importlib import metadata

__title__ = "aiosu"
__author__ = "Nice Aesthetics"
__license__ = "GPLv3+"
__copyright__ = f"Copyright {date.today().year} {__author__}"

from . import events
from . import exceptions
from . import helpers
from . import utils
from . import v1
from . import v2

from . import models  # isort: skip

__all__ = (
    "events",
    "exceptions",
    "helpers",
    "models",
    "utils",
    "v1",
    "v2",
)

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    import toml

    __version__ = toml.load("pyproject.toml")["tool"]["poetry"]["version"] + "dev"



"""
This module contains models for news post objects.
"""

from __future__ import annotations

from datetime import datetime
from typing import Literal
from typing import Optional

from .base import BaseModel
from .common import CursorModel

__all__ = (
    "Navigation",
    "NewsListing",
    "NewsPost",
    "NewsSearch",
    "NewsSortType",
)

NewsSortType = Literal["published_asc", "published_desc"]


class NewsSearch(BaseModel):
    limit: int
    sort: NewsSortType


class Navigation(BaseModel):
    newer: Optional[NewsPost] = None
    older: Optional[NewsPost] = None


class NewsPost(BaseModel):
    id: int
    title: str
    slug: str
    author: str
    edit_url: str
    published_at: datetime
    updated_at: datetime
    first_image: Optional[str] = None
    content: Optional[str] = None
    preview: Optional[str] = None
    navigation: Optional[Navigation] = None


class NewsSidebar(BaseModel):
    current_year: int
    years: list[int]
    news_posts: list[NewsPost]


class NewsListing(CursorModel):
    news_posts: list[NewsPost]
    search: NewsSearch
    news_sidebar: NewsSidebar



"""
This module contains models for User objects.
"""

from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime
from enum import Enum
from enum import unique
from functools import cached_property
from typing import Literal
from typing import Optional

from pydantic import Field
from pydantic import computed_field

from .base import BaseModel
from .base import cast_float
from .base import cast_int
from .common import Country
from .common import HTMLBody
from .common import TimestampedCount
from .gamemode import Gamemode

__all__ = (
    "ManiaStatsVariantsType",
    "User",
    "UserAccountHistory",
    "UserAccountHistoryType",
    "UserBadge",
    "UserGradeCounts",
    "UserGroup",
    "UserKudosu",
    "UserLevel",
    "UserProfileCover",
    "UserProfileTournamentBanner",
    "UserQueryType",
    "UserRankHighest",
    "UserRankHistoryElement",
    "UserRelation",
    "UserStats",
    "UserStatsRulesets",
    "UserStatsVariant",
)


UserAccountHistoryType = Literal[
    "note",
    "restriction",
    "silence",
    "tournament_ban",
]

ManiaStatsVariantsType = Literal[
    "4k",
    "7k",
]

OLD_QUERY_TYPES = {
    "ID": "id",
    "USERNAME": "string",
}


@unique
class UserQueryType(Enum):
    ID = "id"
    USERNAME = "username"

    @computed_field  # type: ignore
    @property
    def old_api_name(self) -> str:
        return OLD_QUERY_TYPES[self.name]

    @computed_field  # type: ignore
    @property
    def new_api_name(self) -> str:
        return self.value

    @classmethod
    def _missing_(cls, query: object) -> UserQueryType:
        if isinstance(query, str):
            query = query.lower()
        for q in list(UserQueryType):
            if query in (q.old_api_name, q.new_api_name):
                return q
        raise ValueError(f"UserQueryType {query} does not exist.")


class UserLevel(BaseModel):
    current: int
    progress: int

    @classmethod
    def _from_api_v1(cls, data: Mapping[str, object]) -> UserLevel:
        level = cast_float(data["level"])
        current = int(level)
        progress = (level - current) * 100
        return cls.model_validate({"current": current, "progress": int(progress)})


class UserKudosu(BaseModel):
    total: int
    available: int


class UserRankHistoryElement(BaseModel):
    mode: str
    data: list[int]

    @computed_field  # type: ignore
    @cached_property
    def average_gain(self) -> float:
        r"""Average rank gain.

        :return: Average rank gain for a user
        :rtype: float
        """
        if not self.data:
            return 0.0
        return (self.data[0] - self.data[-1]) / len(self.data)


class UserRankHighest(BaseModel):
    rank: int
    updated_at: datetime


class UserProfileCover(BaseModel):
    url: str
    custom_url: Optional[str] = None
    id: Optional[str] = None


class UserProfileTournamentBanner(BaseModel):
    tournament_id: int
    id: Optional[int] = None
    image: Optional[str] = None
    image_2_x: Optional[str] = Field(default=None, alias="image@2x")


class UserBadge(BaseModel):
    awarded_at: datetime
    description: str
    image_url: str
    image_2x_url: str = Field(alias="image@2x_url")
    url: str


class UserAccountHistory(BaseModel):
    id: int
    timestamp: datetime
    length: int
    permanent: bool
    type: UserAccountHistoryType
    description: Optional[str] = None


class UserGradeCounts(BaseModel):
    ssh: int
    """Number of Silver SS ranks achieved."""
    ss: int
    """Number of SS ranks achieved."""
    sh: int
    """Number of Silver S ranks achieved."""
    s: int
    """Number of S ranks achieved."""
    a: int
    """Number of A ranks achieved."""

    @classmethod
    def _from_api_v1(cls, data: Mapping[str, object]) -> UserGradeCounts:
        return cls.model_validate(
            {
                "ss": cast_int(data["count_rank_ss"]),
                "ssh": cast_int(data["count_rank_ssh"]),
                "s": cast_int(data["count_rank_s"]),
                "sh": cast_int(data["count_rank_sh"]),
                "a": cast_int(data["count_rank_a"]),
            },
        )


class UserGroup(BaseModel):
    id: int
    identifier: str
    name: str
    short_name: str
    has_listing: bool
    has_playmodes: bool
    is_probationary: bool
    colour: Optional[str] = None
    playmodes: Optional[list[Gamemode]] = None
    description: Optional[str] = None


class UserStatsVariant(BaseModel):
    mode: Gamemode
    variant: str
    pp: float
    country_rank: Optional[int] = None
    global_rank: Optional[int] = None


class UserStats(BaseModel):
    """Fields are marked as optional since they might be missing from rankings other than performance."""

    ranked_score: Optional[int] = None
    play_count: Optional[int] = None
    grade_counts: Optional[UserGradeCounts] = None
    total_hits: Optional[int] = None
    is_ranked: Optional[bool] = None
    total_score: Optional[int] = None
    level: Optional[UserLevel] = None
    hit_accuracy: Optional[float] = None
    play_time: Optional[int] = None
    pp: Optional[float] = None
    pp_exp: Optional[float] = None
    replays_watched_by_others: Optional[int] = None
    maximum_combo: Optional[int] = None
    global_rank: Optional[int] = None
    global_rank_exp: Optional[int] = None
    country_rank: Optional[int] = None
    user: Optional[User] = None
    count_300: Optional[int] = None
    count_100: Optional[int] = None
    count_50: Optional[int] = None
    count_miss: Optional[int] = None
    variants: Optional[list[UserStatsVariant]] = None

    @computed_field  # type: ignore
    @cached_property
    def pp_per_playtime(self) -> float:
        r"""PP per playtime.

        :return: PP per playtime
        :rtype: float
        """
        if not self.play_time or not self.pp:
            return 0.0
        return self.pp / self.play_time * 3600

    @classmethod
    def _from_api_v1(cls, data: Mapping[str, object]) -> UserStats:
        """Some fields can be None, we want to force them to cast to a value."""
        return cls.model_validate(
            {
                "level": UserLevel._from_api_v1(data),
                "pp": cast_float(data["pp_raw"]),
                "global_rank": cast_int(data["pp_rank"]),
                "country_rank": cast_int(data["pp_country_rank"]),
                "ranked_score": cast_int(data["ranked_score"]),
                "hit_accuracy": cast_float(data["accuracy"]),
                "play_count": cast_int(data["playcount"]),
                "play_time": cast_int(data["total_seconds_played"]),
                "total_score": cast_int(data["total_score"]),
                "total_hits": cast_int(data["count300"])
                + cast_int(data["count100"])
                + cast_int(data["count50"]),
                "is_ranked": cast_float(data["pp_raw"]) != 0,
                "grade_counts": UserGradeCounts._from_api_v1(data),
                "count_300": cast_int(data["count300"]),
                "count_100": cast_int(data["count100"]),
                "count_50": cast_int(data["count50"]),
            },
        )


class UserStatsRulesets(BaseModel):
    osu: Optional[UserStats] = None
    taiko: Optional[UserStats] = None
    fruits: Optional[UserStats] = None
    mania: Optional[UserStats] = None


class UserAchievmement(BaseModel):
    achieved_at: datetime
    achievement_id: int


class UserRelation(BaseModel):
    target_id: int
    relation_type: str
    mutual: bool
    target: Optional[User] = None


class User(BaseModel):
    avatar_url: str
    country_code: str
    id: int
    username: str
    default_group: Optional[str] = None
    is_active: Optional[bool] = None
    is_bot: Optional[bool] = None
    is_online: Optional[bool] = None
    is_supporter: Optional[bool] = None
    pm_friends_only: Optional[bool] = None
    profile_colour: Optional[str] = None
    is_deleted: Optional[bool] = None
    last_visit: Optional[datetime] = None
    discord: Optional[str] = None
    has_supported: Optional[bool] = None
    interests: Optional[str] = None
    join_date: Optional[datetime] = None
    kudosu: Optional[UserKudosu] = None
    location: Optional[str] = None
    max_blocks: Optional[int] = None
    max_friends: Optional[int] = None
    occupation: Optional[str] = None
    playmode: Optional[Gamemode] = None
    playstyle: Optional[list[str]] = None
    post_count: Optional[int] = None
    profile_order: Optional[list[str]] = None
    title: Optional[str] = None
    twitter: Optional[str] = None
    website: Optional[str] = None
    country: Optional[Country] = None
    cover: Optional[UserProfileCover] = None
    is_restricted: Optional[bool] = None
    account_history: Optional[list[UserAccountHistory]] = None
    active_tournament_banners: Optional[list[UserProfileTournamentBanner]] = None
    badges: Optional[list[UserBadge]] = None
    beatmap_playcounts_count: Optional[int] = None
    favourite_beatmapset_count: Optional[int] = None
    follow_user_mapping: Optional[list[int]] = None
    follower_count: Optional[int] = None
    friends: Optional[list[UserRelation]] = None
    graveyard_beatmapset_count: Optional[int] = None
    groups: Optional[list[UserGroup]] = None
    loved_beatmapset_count: Optional[int] = None
    mapping_follower_count: Optional[int] = None
    monthly_playcounts: Optional[list[TimestampedCount]] = None
    page: Optional[HTMLBody] = None
    pending_beatmapset_count: Optional[int] = None
    previous_usernames: Optional[list[str]] = None
    rank_highest: Optional[UserRankHighest] = None
    rank_history: Optional[UserRankHistoryElement] = None
    ranked_beatmapset_count: Optional[int] = None
    replays_watched_counts: Optional[list[TimestampedCount]] = None
    scores_best_count: Optional[int] = None
    scores_first_count: Optional[int] = None
    scores_recent_count: Optional[int] = None
    statistics: Optional[UserStats] = None
    statistics_rulesets: Optional[UserStatsRulesets] = None
    support_level: Optional[int] = None
    unread_pm_count: Optional[int] = None
    user_achievements: Optional[list[UserAchievmement]] = None

    @computed_field  # type: ignore
    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/users/{self.id}"

    @classmethod
    def _from_api_v1(cls, data: Mapping[str, object]) -> User:
        return cls.model_validate(
            {
                "avatar_url": f"https://s.ppy.sh/a/{data['user_id']}",
                "country_code": data["country"],
                "id": data["user_id"],
                "username": data["username"],
                "join_date": data["join_date"],
                "statistics": UserStats._from_api_v1(data),
            },
        )



"""
This module contains models for API v2 token objects.
"""

from __future__ import annotations

from datetime import datetime
from datetime import timedelta
from functools import cached_property

import jwt
from pydantic import computed_field
from pydantic import model_validator

from .base import FrozenModel
from .scopes import Scopes

__all__ = ("OAuthToken",)


class OAuthToken(FrozenModel):
    token_type: str = "Bearer"
    """Defaults to 'Bearer'"""
    access_token: str = ""
    refresh_token: str = ""
    expires_on: datetime = datetime.utcfromtimestamp(31536000)
    """Can be a datetime.datetime object or a string. Alternatively, expires_in may be passed representing the number of seconds the token will be valid for."""

    @computed_field  # type: ignore
    @cached_property
    def owner_id(self) -> int:
        if not self.access_token:
            return 0
        decoded = jwt.decode(self.access_token, options={"verify_signature": False})
        if decoded["sub"]:
            return int(decoded["sub"])
        return 0

    @computed_field  # type: ignore
    @cached_property
    def scopes(self) -> Scopes:
        if not self.access_token:
            return Scopes.PUBLIC
        decoded = jwt.decode(self.access_token, options={"verify_signature": False})
        return Scopes.from_api_list(decoded["scopes"])

    @computed_field  # type: ignore
    @cached_property
    def can_refresh(self) -> bool:
        """Returns True if the token can be refreshed."""
        return bool(self.refresh_token)

    @model_validator(mode="before")
    @classmethod
    def _set_expires_on(cls, values: dict[str, object]) -> dict[str, object]:
        expires_in = values.get("expires_in")
        if isinstance(expires_in, int):
            values["expires_on"] = datetime.utcnow() + timedelta(
                seconds=expires_in,
            )
        return values



"""
This module contains models for mods.
"""

from __future__ import annotations

from collections import UserList
from enum import IntEnum
from enum import unique
from functools import reduce
from typing import TYPE_CHECKING

from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema
from pydantic_core import core_schema

if TYPE_CHECKING:
    from typing import Union

__all__ = (
    "FreemodAllowed",
    "KeyMod",
    "Mod",
    "Mods",
    "ScoreIncreaseMods",
    "SpeedChangingMods",
)

_mod_short_names = {
    "NoMod": "NM",
    "NoFail": "NF",
    "Easy": "EZ",
    "TouchDevice": "TD",
    "Hidden": "HD",
    "HardRock": "HR",
    "SuddenDeath": "SD",
    "DoubleTime": "DT",
    "Relax": "RX",
    "HalfTime": "HT",
    "Nightcore": "NC",
    "Flashlight": "FL",
    "Autoplay": "AT",
    "SpunOut": "SO",
    "Autopilot": "AP",
    "Perfect": "PF",
    "Key4": "4K",
    "Key5": "5K",
    "Key6": "6K",
    "Key7": "7K",
    "Key8": "8K",
    "FadeIn": "FI",
    "Random": "RD",
    "Cinema": "CN",
    "Target": "TP",
    "Key9": "9K",
    "KeyCoop": "CO",
    "Key1": "1K",
    "Key3": "3K",
    "Key2": "2K",
    "ScoreV2": "V2",
    "Mirror": "MR",
}


@unique
class Mod(IntEnum):
    """Bitwise Flags representing osu! mods."""

    NoMod = 0
    NoFail = 1 << 0
    Easy = 1 << 1
    TouchDevice = 1 << 2
    Hidden = 1 << 3
    HardRock = 1 << 4
    SuddenDeath = 1 << 5
    DoubleTime = 1 << 6
    Relax = 1 << 7
    HalfTime = 1 << 8
    Nightcore = 1 << 9
    """Only set along with DoubleTime. i.e: NC only gives 576"""
    Flashlight = 1 << 10
    Autoplay = 1 << 11
    SpunOut = 1 << 12
    Autopilot = 1 << 13
    """Called Relax2 on osu! API documentation"""
    Perfect = 1 << 14
    """Only set along with SuddenDeath. i.e: PF only gives 16416"""
    Key4 = 1 << 15
    Key5 = 1 << 16
    Key6 = 1 << 17
    Key7 = 1 << 18
    Key8 = 1 << 19
    FadeIn = 1 << 20
    Random = 1 << 21
    Cinema = 1 << 22
    Target = 1 << 23
    Key9 = 1 << 24
    KeyCoop = 1 << 25
    Key1 = 1 << 26
    Key3 = 1 << 27
    Key2 = 1 << 28
    ScoreV2 = 1 << 29
    Mirror = 1 << 30

    @property
    def bitmask(self) -> int:
        return self.value

    @property
    def short_name(self) -> str:
        return _mod_short_names[self.name]

    def __str__(self) -> str:
        return self.short_name

    @classmethod
    def from_type(cls, __o: object) -> Mod:
        """Get a Mod from a string or int.

        :param __o: The string or int to get the Mod from
        :type __o: object
        :return: The Mod
        :rtype: Mod
        :raises ValueError: If the Mod does not exist
        """
        from .lazer import LazerMod  # Lazy import to avoid circular imports

        if isinstance(__o, cls):
            return __o
        if isinstance(__o, LazerMod):  # Attempt lossy conversion to stable mods
            try:
                return cls.from_type(__o.acronym)
            except ValueError:
                return cls.NoMod
        for mod in list(Mod):
            if __o == mod.short_name or __o == mod.bitmask:
                return mod
        raise ValueError(f"Mod {__o!r} does not exist.")

    @classmethod
    def _missing_(cls, query: object) -> Mod:
        return cls.from_type(query)


class Mods(UserList):
    """List of Mod objects"""

    def __init__(self, mods: Union[list[str], str, int]) -> None:
        super().__init__(self)
        self.data = []
        if isinstance(mods, str):  # string of mods
            mods = [mods[i : i + 2] for i in range(0, len(mods), 2)]

        if isinstance(mods, int):  # Bitwise representation of mods
            self.data = [mod for mod in list(Mod) if mod & mods]
        elif isinstance(mods, list) or isinstance(mods, Mods):  # List of Mod types
            self.data = [Mod(mod) for mod in mods]  # type: ignore
        else:
            raise TypeError(
                f"Mods must be a list of Mod types, a string, or an int. Not {type(mods)}",
            )

        if Mod.Nightcore in self and Mod.DoubleTime not in self:
            self.data.append(Mod.DoubleTime)
        if Mod.Perfect in self and Mod.SuddenDeath not in self:
            self.data.append(Mod.SuddenDeath)

    @property
    def bitwise(self) -> int:
        r"""Bitwise representation.

        :return: Bitwise representation of the mod combination
        :rtype: int
        """
        return reduce(lambda x, y: int(x) | int(y), self, 0)

    def __str__(self) -> str:
        if len(self) == 0:
            return "NM"

        result: str = ""
        for mod in self:
            if Mod.Nightcore in self and mod is Mod.DoubleTime:
                continue
            if Mod.Perfect in self and mod is Mod.SuddenDeath:
                continue

            result += mod.short_name
        return result

    def __int__(self) -> int:
        return self.bitwise

    def __and__(self, __o: object) -> int:
        if isinstance(__o, (int, Mod, Mods)):
            return int(self) & int(__o)

        return NotImplemented

    def __or__(self, __o: object) -> int:
        if isinstance(__o, (int, Mod, Mods)):
            return int(self) | int(__o)

        return NotImplemented

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: type[object],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        return core_schema.no_info_before_validator_function(
            cls,
            core_schema.json_or_python_schema(
                json_schema=core_schema.list_schema(),
                python_schema=handler(source_type),
            ),
            serialization=core_schema.to_string_ser_schema(when_used="json"),
        )


KeyMod = (
    Mod.Key1
    | Mod.Key2
    | Mod.Key3
    | Mod.Key4
    | Mod.Key5
    | Mod.Key6
    | Mod.Key7
    | Mod.Key8
    | Mod.Key9
    | Mod.KeyCoop
)
FreemodAllowed = (
    Mod.NoFail
    | Mod.Easy
    | Mod.Hidden
    | Mod.HardRock
    | Mod.SuddenDeath
    | Mod.Flashlight
    | Mod.Relax
    | Mod.SpunOut
    | Mod.Autopilot
    | Mod.Perfect
    | Mod.Key4
    | Mod.Key5
    | Mod.Key6
    | Mod.Key7
    | Mod.Key8
    | Mod.FadeIn
    | Mod.Random
    | Mod.Key9
    | Mod.KeyCoop
    | Mod.Key1
    | Mod.Key3
    | Mod.Key2
    | Mod.Mirror
)
ScoreIncreaseMods = Mod.Hidden | Mod.HardRock | Mod.DoubleTime | Mod.Flashlight
SpeedChangingMods = Mod.DoubleTime | Mod.HalfTime | Mod.Nightcore



"""
This module contains models for chat.
"""

from __future__ import annotations

from typing import Literal
from typing import Optional

from pydantic import Field

from .base import BaseModel
from .common import CurrentUserAttributes
from .user import User

__all__ = (
    "ChatChannel",
    "ChatChannelResponse",
    "ChatChannelType",
    "ChatIncludeType",
    "ChatMessage",
    "ChatMessageCreateResponse",
    "ChatUpdateResponse",
    "ChatUserSilence",
)

ChatChannelType = Literal[
    "PM",
    "PUBLIC",
    "PRIVATE",
    "MULTIPLAYER",
    "SPECTATOR",
    "TEMPORARY",
    "GROUP",
    "ANNOUNCE",
]

ChatIncludeType = Literal[
    "messages",
    "presence",
    "silences",
]


class ChatUserSilence(BaseModel):
    id: int
    user_id: int


class ChatChannel(BaseModel):
    id: int = Field(alias="channel_id")
    type: ChatChannelType
    name: str
    moderated: bool
    message_length_limit: int
    icon: Optional[str] = None
    description: Optional[str] = None
    last_message_id: Optional[int] = None
    user_ids: Optional[list[int]] = Field(default=None, alias="users")
    current_user_attributes: Optional[CurrentUserAttributes] = None


class ChatMessage(BaseModel):
    message_id: int
    sender_id: int
    channel_id: int
    timestamp: str
    content: str
    is_action: bool
    uuid: Optional[str] = None
    sender: Optional[User] = None


class ChatMessageCreateResponse(BaseModel):
    channel: ChatChannel
    message: ChatMessage


class ChatUpdateResponse(BaseModel):
    messages: Optional[list[ChatMessage]] = None
    presence: Optional[list[ChatChannel]] = None
    silences: list


class ChatChannelResponse(BaseModel):
    channel: ChatChannel
    users: Optional[list[User]] = None



"""
This module contains base models for objects.
"""

from __future__ import annotations

from typing import SupportsFloat
from typing import SupportsInt

import pydantic
from pydantic import ConfigDict

__all__ = (
    "BaseModel",
    "FrozenModel",
)


class BaseModel(pydantic.BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    @classmethod
    def model_validate_file(cls, path: str) -> BaseModel:
        """Validates a model from a file.

        :param path: The path to the file
        :type path: str
        :raises TypeError: If the file is not a JSON file
        :return: The validated model
        :rtype: aiosu.models.base.BaseModel
        """
        with open(path) as f:
            return cls.model_validate_json(f.read())


class FrozenModel(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        populate_by_name=True,
        frozen=True,
    )


def cast_int(v: object) -> int:
    if v is None:
        return 0
    if isinstance(v, (SupportsInt, str)):
        return int(v)

    raise ValueError(f"{v} is not a valid value.")


def cast_float(v: object) -> float:
    if v is None:
        return 0.0
    if isinstance(v, (SupportsFloat, str)):
        return float(v)

    raise ValueError(f"{v} is not a valid value.")



"""
This module contains models for seasonal background objects.
"""

from __future__ import annotations

from datetime import datetime

from .base import BaseModel
from .user import User

__all__ = (
    "Background",
    "SeasonalBackgroundSet",
)


class Background(BaseModel):
    url: str
    user: User


class SeasonalBackgroundSet(BaseModel):
    ends_at: datetime
    backgrounds: list[Background]



"""
This module contains models for API events.
"""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from typing import Literal
from typing import Optional

from pydantic import Field

from .base import BaseModel
from .beatmap import BeatmapRankStatus
from .common import Achievement
from .common import CursorModel
from .gamemode import Gamemode

if TYPE_CHECKING:
    pass

__all__ = (
    "Event",
    "EventBeatmap",
    "EventBeatmapset",
    "EventResponse",
    "EventType",
    "EventUser",
)

EventType = Literal[
    "achievement",
    "beatmapPlaycount",
    "beatmapsetApprove",
    "beatmapsetDelete",
    "beatmapsetRevive",
    "beatmapsetUpdate",
    "beatmapsetUpload",
    "rank",
    "rankLost",
    "usernameChange",
    "userSupportAgain",
    "userSupportFirst",
    "userSupportGift",
]


class EventBeatmap(BaseModel):
    title: str
    url: str


class EventBeatmapset(BaseModel):
    title: str
    url: str


class EventUser(BaseModel):
    username: str
    url: str
    previous_username: Optional[str] = Field(default=None, alias="previousUsername")


class Event(BaseModel):
    created_at: datetime
    id: int
    type: EventType
    r"""Information on types: https://github.com/ppy/osu-web/blob/master/resources/js/interfaces/event-json.ts"""
    parse_error: Optional[bool] = None
    achievement: Optional[Achievement] = None
    user: Optional[EventUser] = None
    beatmap: Optional[EventBeatmap] = None
    beatmapset: Optional[EventBeatmapset] = None
    approval: Optional[BeatmapRankStatus] = None
    count: Optional[int] = None
    rank: Optional[int] = None
    mode: Optional[Gamemode] = None
    score_rank: Optional[str] = Field(default=None, alias="scoreRank")


class EventResponse(CursorModel):
    events: list[Event]



"""
This module contains models for multiplayer.
"""

from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime
from typing import Literal
from typing import Optional
from typing import Union

from pydantic import Field
from pydantic import model_validator

from .base import BaseModel
from .beatmap import Beatmap
from .common import CursorModel
from .gamemode import Gamemode
from .lazer import LazerMod
from .lazer import LazerScore
from .mods import Mods
from .score import Score
from .user import User

__all__ = (
    "MultiplayerEvent",
    "MultiplayerEventType",
    "MultiplayerGame",
    "MultiplayerLeaderboardItem",
    "MultiplayerLeaderboardResponse",
    "MultiplayerMatch",
    "MultiplayerMatchResponse",
    "MultiplayerMatchesResponse",
    "MultiplayerQueueMode",
    "MultiplayerRoom",
    "MultiplayerRoomCategory",
    "MultiplayerRoomGroupType",
    "MultiplayerRoomMode",
    "MultiplayerRoomsResponse",
    "MultiplayerScoreSortType",
    "MultiplayerScoresAround",
    "MultiplayerScoresResponse",
    "MultiplayerScoringType",
    "MultiplayerTeamType",
)

MultiplayerScoringType = Literal["score", "accuracy", "combo", "scorev2"]
MultiplayerTeamType = Literal["head-to-head", "tag-coop", "team-vs", "tag-team-vs"]
MultiplayerScoreSortType = Literal["score_asc", "score_desc"]
MultiplayerEventType = Literal[
    "match-created",
    "match-disbanded",
    "host-changed",
    "player-joined",
    "player-left",
    "player-kicked",
    "match-created-no-user",
    "match-disbanded-no-user",
    "host-changed-no-user",
    "player-joined-no-user",
    "player-left-no-user",
    "player-kicked-no-user",
    "other",
]
MultiplayerRoomMode = Literal["owned", "participated", "ended"]
MultiplayerRoomCategory = Literal["normal", "spotlight", "featured_artist"]
MultiplayerRoomGroupType = Literal["playlists", "realtime"]
MultiplayerQueueMode = Literal["host_only", "all_players", "all_players_round_robin"]


class MultiplayerScoresAround(BaseModel):
    higher: MultiplayerScoresResponse
    lower: MultiplayerScoresResponse


class MultiplayerScoresResponse(CursorModel):
    scores: list[LazerScore]
    user_score: Optional[LazerScore] = None
    total: Optional[int] = None


class MultiplayerMatch(BaseModel):
    id: int
    name: str
    start_time: datetime
    end_time: Optional[datetime] = None


class MultiplayerGame(BaseModel):
    id: int
    start_time: datetime
    mode: Gamemode
    scoring_type: MultiplayerScoringType
    team_type: MultiplayerTeamType
    mods: list[Union[Mods, LazerMod]]
    beatmap_id: int
    scores: list[Score]
    beatmap: Optional[Beatmap] = None
    end_time: Optional[datetime] = None


class MultiplayerEvent(BaseModel):
    id: int
    timestamp: datetime
    type: MultiplayerEventType
    user_id: Optional[int] = None
    game: Optional[MultiplayerGame] = None

    @model_validator(mode="before")
    @classmethod
    def _set_type(cls, values: dict[str, object]) -> dict[str, object]:
        if "detail" in values and isinstance(values["detail"], Mapping):
            values["type"] = values["detail"]["type"]
        return values


class MultiplayerMatchResponse(BaseModel):
    match: MultiplayerMatch
    events: list[MultiplayerEvent]
    users: list[User]
    first_event_id: int
    latest_event_id: int
    current_game_id: Optional[int] = None


class MultiplayerMatchesResponse(CursorModel):
    matches: list[MultiplayerMatch]


class MultiplayerPlaylistItem(BaseModel):
    id: int
    room_id: int
    beatmap_id: int
    mode: Gamemode = Field(alias="ruleset_id")
    allowed_mods: list[LazerMod]
    required_mods: list[LazerMod]
    expired: bool
    owner_id: int
    beatmap: Beatmap
    playlist_order: Optional[int] = None
    played_at: Optional[datetime] = None


class MultiplayerRoom(BaseModel):
    id: int
    name: str
    category: MultiplayerRoomCategory
    type: MultiplayerRoomGroupType
    user_id: int
    channel_id: int
    active: bool
    has_password: bool
    auto_skip: bool
    host: User
    queue_mode: MultiplayerQueueMode
    playlist: list[MultiplayerPlaylistItem]
    recent_participants: list[User]
    participant_count: Optional[int] = None
    starts_at: Optional[datetime] = None
    ends_at: Optional[datetime] = None
    max_attempts: Optional[int] = None


class MultiplayerLeaderboardItem(BaseModel):
    accuracy: float
    attempts: int
    completed: int
    pp: float
    room_id: int
    total_score: int
    user_id: int
    user: User
    position: Optional[int] = None


class MultiplayerLeaderboardResponse(BaseModel):
    leaderboard: list[MultiplayerLeaderboardItem]
    user_score: Optional[MultiplayerLeaderboardItem] = None


class MultiplayerRoomsResponse(CursorModel):
    """Currently unused. Relevant for api-version >= 99999999"""

    rooms: list[MultiplayerRoom]



"""
This module contains custom exception types.
"""

from __future__ import annotations

__all__ = ("APIException", "InvalidClientRequestedError", "RefreshTokenExpiredError")


class APIException(Exception):
    """API Exception Class

    :param status: status code from the API
    :type status: int
    :param message: error message returned, defaults to ""
    :type message: str, optional
    """

    def __init__(self, status: int, message: str = "") -> None:
        super().__init__(message)
        self.status = status


class RefreshTokenExpiredError(Exception):
    """Refresh Token Expired Error

    :param message: error message, defaults to ""
    :type message: str, optional
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class InvalidClientRequestedError(Exception):
    """Invalid Client Requested Error

    :param message: error message, defaults to ""
    :type message: str, optional
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)



"""
This module contains models for featured artists.
"""

from __future__ import annotations

from datetime import datetime
from typing import Literal
from typing import Optional

from pydantic import Field

from .base import BaseModel
from .beatmap import Beatmapset
from .common import CursorModel

__all__ = (
    "Artist",
    "ArtistAlbum",
    "ArtistLabel",
    "ArtistSearch",
    "ArtistSortType",
    "ArtistTrack",
    "ArtistTracksResponse",
)

ArtistSortType = Literal[
    "album_asc",
    "album_desc",
    "artist_asc",
    "artist_desc",
    "bpm_asc",
    "bpm_desc",
    "genre_asc",
    "genre_desc",
    "length_asc",
    "length_desc",
    "relevance_asc",
    "relevance_desc",
    "title_asc",
    "title_desc",
    "update_asc",
    "update_desc",
]


class ArtistSearch(BaseModel):
    is_default_sort: bool
    sort: ArtistSortType


class ArtistLabel(BaseModel):
    id: int
    artists: list[Artist]
    name: str
    description: str
    icon_url: str
    header_url: str
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    soundcloud: Optional[str] = None
    website: Optional[str] = None


class Artist(BaseModel):
    id: int
    name: str
    user_id: Optional[int] = None
    description: Optional[str] = None
    visible: Optional[bool] = None
    header_url: Optional[str] = None
    cover_url: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    beatmapsets: Optional[list[Beatmapset]] = None
    label: Optional[ArtistLabel] = None
    albums: Optional[list[ArtistAlbum]] = None
    tracks: Optional[list[ArtistTrack]] = None
    bandcamp: Optional[str] = None
    facebook: Optional[str] = None
    patreon: Optional[str] = None
    soundcloud: Optional[str] = None
    spotify: Optional[str] = None
    twitter: Optional[str] = None
    website: Optional[str] = None
    youtube: Optional[str] = None
    video_url: Optional[str] = None


class ArtistAlbum(BaseModel):
    id: int
    artist_id: int
    title: str
    title_romanized: str
    genre: str
    is_new: bool
    cover_url: str


class ArtistTrack(BaseModel):
    id: int
    artist_id: int
    bpm: float
    cover_url: str
    exclusive: bool
    genre: str
    is_new: bool
    length: str
    title: str
    artist: Artist
    osz_url: str = Field(alias="osz")
    preview_url: str = Field(alias="preview")
    album: Optional[ArtistAlbum] = None
    album_id: Optional[int] = None
    updated_at: Optional[datetime] = None
    version: Optional[str] = None


class ArtistTracksResponse(CursorModel):
    """Artist response model."""

    artist_tracks: list[ArtistTrack]
    search: Optional[ArtistSearch] = None



"""
This module contains a collection of utility functions.
"""

# isort: dont-add-imports

from . import auth
from . import binary
from . import performance
from . import replay

from . import accuracy  # isort: skip



"""
This module contains models for miscellaneous objects.
"""

from __future__ import annotations

from collections.abc import Coroutine
from datetime import datetime
from functools import cached_property
from functools import partial
from typing import Literal
from typing import Optional

from emojiflags.lookup import lookup as flag_lookup  # type: ignore
from pydantic import Field
from pydantic import computed_field
from pydantic import field_validator

from .base import BaseModel
from .gamemode import Gamemode

__all__ = (
    "Achievement",
    "BeatmapScoreboardType",
    "Country",
    "CurrentUserAttributes",
    "CursorModel",
    "HTMLBody",
    "ScoreType",
    "SortType",
    "TimestampedCount",
)


SortType = Literal["id_asc", "id_desc"]
ScoreType = Literal[
    "solo_score",
    "score_best_osu",
    "score_best_taiko",
    "score_best_fruits",
    "score_best_mania",
    "score_osu",
    "score_taiko",
    "score_fruits",
    "score_mania",
    "legacy_match_score",
]
BeatmapScoreboardType = Literal["global", "country", "friend"]


class TimestampedCount(BaseModel):
    start_date: datetime
    count: int

    @field_validator("start_date", mode="before")
    @classmethod
    def _date_validate(cls, v: object) -> datetime:
        if isinstance(v, str):
            return datetime.strptime(v, "%Y-%m-%d")
        if isinstance(v, datetime):
            return v

        raise ValueError(f"{v} is not a valid value.")


class Achievement(BaseModel):
    id: int
    name: str
    slug: str
    description: str
    grouping: str
    icon_url: str
    ordering: int
    mode: Optional[Gamemode] = None
    instructions: Optional[str] = None


class Country(BaseModel):
    code: str
    name: str

    @computed_field  # type: ignore
    @cached_property
    def flag_emoji(self) -> str:
        r"""Emoji for the flag.

        :return: Unicode emoji representation of the country's flag
        :rtype: str
        """
        return flag_lookup(self.code)


class HTMLBody(BaseModel):
    html: str
    raw: Optional[str] = None
    bbcode: Optional[str] = None


class PinAttributes(BaseModel):
    is_pinned: bool
    score_id: int
    score_type: ScoreType


class CurrentUserAttributes(BaseModel):
    can_beatmap_update_owner: Optional[bool] = None
    can_delete: Optional[bool] = None
    can_edit_metadata: Optional[bool] = None
    can_edit_tags: Optional[bool] = None
    can_hype: Optional[bool] = None
    can_hype_reason: Optional[str] = None
    can_love: Optional[bool] = None
    can_remove_from_loved: Optional[bool] = None
    is_watching: Optional[bool] = None
    new_hype_time: Optional[datetime] = None
    nomination_modes: Optional[list[Gamemode]] = None
    remaining_hype: Optional[int] = None
    can_destroy: Optional[bool] = None
    can_reopen: Optional[bool] = None
    can_moderate_kudosu: Optional[bool] = None
    can_resolve: Optional[bool] = None
    vote_score: Optional[int] = None
    can_message: Optional[bool] = None
    can_message_error: Optional[str] = None
    last_read_id: Optional[int] = None
    can_new_comment: Optional[bool] = None
    can_new_comment_reason: Optional[str] = None
    pin: Optional[PinAttributes] = None


class CursorModel(BaseModel):
    r"""NOTE: This model is not serializable by orjson directly.

    Use the provided .model_dump_json() or .model_dump() methods instead.
    """

    cursor_string: Optional[str] = None
    next: Optional[partial[Coroutine[object, object, CursorModel]]] = Field(
        default=None,
        exclude=True,
    )
    """Partial function to get the next page of results."""



"""
This module contains API models.
"""

# isort: dont-add-imports

from .artist import *
from .backgrounds import *
from .base import *
from .beatmap import *
from .changelog import *
from .chat import *
from .comment import *
from .common import *
from .event import *
from .files import *
from .forum import *
from .gamemode import *
from .kudosu import *
from .lazer import *
from .mods import *
from .multiplayer import *
from .news import *
from .oauthtoken import *
from .performance import *
from .rankings import *
from .scopes import *
from .score import *
from .search import *
from .spotlight import *
from .user import *
from .wiki import *



"""
This module contains models for wiki objects.
"""

from __future__ import annotations

from typing import Optional

from .base import BaseModel

__all__ = ("WikiPage",)


class WikiPage(BaseModel):
    title: str
    path: str
    locale: str
    available_locales: list[str]
    layout: str
    markdown: str
    subtitle: Optional[str] = None
    tags: Optional[list[str]] = None



from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime
from enum import IntEnum
from enum import unique
from typing import Optional

from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator

from ..base import BaseModel
from ..base import cast_int
from ..gamemode import Gamemode
from ..mods import Mods
from ..score import ScoreStatistics

__all__ = (
    "Match",
    "MatchGame",
    "MatchScore",
    "MatchScoringType",
    "MatchTeam",
    "MatchTeamType",
)


@unique
class MatchTeam(IntEnum):
    NONE = 0
    BLUE = 1
    RED = 2


@unique
class MatchScoringType(IntEnum):
    SCORE = 0
    ACCURACY = 1
    COMBO = 2
    SCOREV2 = 3


@unique
class MatchTeamType(IntEnum):
    HEADTOHEAD = 0
    COOP = 1
    TEAMVS = 2
    TAGTEAMVS = 3


class MatchScore(BaseModel):
    slot: int
    team: MatchTeam
    user_id: int
    score: int
    max_combo: int = Field(alias="maxcombo")
    passed: bool = Field(alias="pass")
    perfect: bool
    enabled_mods: Mods
    statistics: ScoreStatistics

    def get_full_mods(self, game: MatchGame) -> Mods:
        r"""Get the mods including globals from the match.

        :param game: The match the score took place in
        :type game: MatchGame
        :return: Mods list of Mod objects
        :rtype: Mods
        """
        return Mods(self.enabled_mods | game.mods)

    @model_validator(mode="before")
    @classmethod
    def _set_statistics(cls, values: dict[str, object]) -> dict[str, object]:
        values["statistics"] = ScoreStatistics._from_api_v1(values)
        return values

    @field_validator("enabled_mods", mode="before")
    @classmethod
    def _set_enabled_mods(cls, v: object) -> int:
        return cast_int(v)

    @field_validator("team", mode="before")
    @classmethod
    def _set_team(cls, v: object) -> int:
        return cast_int(v)


class MatchGame(BaseModel):
    """Multiplayer game API object."""

    id: int = Field(alias="game_id")
    start_time: datetime
    beatmap_id: int
    mode: Gamemode = Field(alias="play_mode")
    match_type: int
    scoring_type: MatchScoringType
    team_type: MatchTeamType
    scores: list[MatchScore]
    mods: Mods
    end_time: Optional[datetime] = None
    """None if game was aborted."""

    @field_validator("mode", mode="before")
    @classmethod
    def _set_mode(cls, v: object) -> int:
        return cast_int(v)

    @field_validator("mods", mode="before")
    @classmethod
    def _set_mods(cls, v: object) -> int:
        return cast_int(v)

    @field_validator("scoring_type", mode="before")
    @classmethod
    def _set_scoring_type(cls, v: object) -> int:
        return cast_int(v)

    @field_validator("team_type", mode="before")
    @classmethod
    def _set_team_type(cls, v: object) -> int:
        return cast_int(v)


class Match(BaseModel):
    """Multiplayer match API object."""

    id: int = Field(alias="match_id")
    name: str
    start_time: datetime
    games: list[MatchGame]
    end_time: Optional[datetime] = None
    """None if game is ongoing."""

    @model_validator(mode="before")
    @classmethod
    def _format_values(cls, values: dict[str, object]) -> dict[str, object]:
        match = values["match"]
        if not isinstance(match, Mapping):
            raise ValueError(f"Invalid match type: {type(match)}")

        return {**match, "games": values["games"]}



"""
This module contains models for comment objects.
"""

from __future__ import annotations

from datetime import datetime
from typing import Literal
from typing import Optional

from .base import BaseModel
from .common import CurrentUserAttributes
from .common import CursorModel
from .user import User

__all__ = (
    "Comment",
    "CommentBundle",
    "CommentSortType",
    "Commentable",
    "CommentableType",
)

CommentSortType = Literal["new", "old", "top"]
CommentableType = Literal["beatmapset", "news_post", "build"]


class Commentable(BaseModel):
    id: Optional[int] = None
    title: Optional[str] = None
    url: Optional[str] = None
    type: Optional[str] = None
    owner_id: Optional[int] = None
    owner_title: Optional[str] = None
    current_user_attributes: Optional[CurrentUserAttributes] = None


class Comment(BaseModel):
    id: int
    commentable_id: int
    commentable_type: CommentableType
    created_at: datetime
    updated_at: datetime
    pinned: bool
    votes_count: int
    replies_count: int
    message: Optional[str] = None
    message_html: Optional[str] = None
    deleted_at: Optional[datetime] = None
    edited_at: Optional[datetime] = None
    edited_by_id: Optional[int] = None
    parent_id: Optional[int] = None
    legacy_name: Optional[str] = None
    user_id: Optional[int] = None


class CommentBundle(CursorModel):
    commentable_meta: list[Commentable]
    comments: list[Comment]
    has_more: bool
    included_comments: list[Comment]
    sort: CommentSortType
    user_follow: bool
    user_votes: list[int]
    users: list[User]
    pinned_comments: Optional[list[Comment]] = None
    total: Optional[int] = None
    top_level_count: Optional[int] = None
    has_more_id: Optional[int] = None



"""
This module contains models for lazer specific data.
"""

from __future__ import annotations

from datetime import datetime
from functools import cached_property
from typing import Optional

from pydantic import Field
from pydantic import computed_field
from pydantic import model_validator

from .base import BaseModel
from .beatmap import Beatmap
from .beatmap import Beatmapset
from .common import CurrentUserAttributes
from .common import ScoreType
from .gamemode import Gamemode
from .score import ScoreWeight
from .user import User

__all__ = (
    "LazerMod",
    "LazerReplayData",
    "LazerScore",
    "LazerScoreStatistics",
)


def calculate_score_completion(
    statistics: LazerScoreStatistics,
    beatmap: Beatmap,
) -> Optional[float]:
    """Calculates completion for a score.

    :param statistics: The statistics of the score
    :type statistics: aiosu.models.lazer.LazerScoreStatistics
    :param beatmap: The beatmap of the score
    :type beatmap: aiosu.models.beatmap.Beatmap
    :raises ValueError: If the gamemode is unknown
    :return: Completion for the given score
    :rtype: Optional[float]
    """
    if not beatmap.count_objects:
        return None

    return (
        (
            statistics.perfect
            + statistics.good
            + statistics.great
            + statistics.ok
            + statistics.meh
            + statistics.miss
        )
        / beatmap.count_objects
    ) * 100


class LazerMod(BaseModel):
    """Temporary model for lazer mods."""

    acronym: str
    settings: dict[str, object] = Field(default_factory=dict)

    def __str__(self) -> str:
        return self.acronym


class LazerScoreStatistics(BaseModel):
    """Relevant information:

    https://github.com/ppy/osu/blob/master/osu.Game/Rulesets/Scoring/HitResult.cs
    https://github.com/ppy/osu-web/blob/master/resources/js/interfaces/solo-score-json.ts
    """

    miss: int = 0
    meh: int = 0
    ok: int = 0
    good: int = 0
    great: int = 0
    perfect: int = 0
    small_tick_miss: int = 0
    small_tick_hit: int = 0
    large_tick_miss: int = 0
    large_tick_hit: int = 0
    small_bonus: int = 0
    large_bonus: int = 0
    ignore_miss: int = 0
    ignore_hit: int = 0
    combo_break: int = 0
    slider_tail_hit: int = 0
    legacy_combo_increase: int = 0

    @property
    def count_300(self) -> int:
        return self.great

    @property
    def count_100(self) -> int:
        return self.ok

    @property
    def count_50(self) -> int:
        return self.meh

    @property
    def count_miss(self) -> int:
        return self.miss

    @property
    def count_geki(self) -> int:
        return self.perfect

    @property
    def count_katu(self) -> int:
        return self.good


class LazerReplayData(BaseModel):
    mods: list[LazerMod]
    statistics: LazerScoreStatistics
    maximum_statistics: LazerScoreStatistics
    online_id: Optional[int] = None
    client_version: Optional[str] = None


class LazerScore(BaseModel):
    accuracy: float
    beatmap_id: int
    ended_at: datetime
    has_replay: bool = Field(validation_alias="replay")
    is_perfect_combo: bool
    legacy_perfect: bool
    max_combo: int
    maximum_statistics: LazerScoreStatistics
    mods: list[LazerMod]
    passed: bool
    rank: str
    ruleset_id: int
    statistics: LazerScoreStatistics
    total_score: int = Field(validation_alias="score")
    type: ScoreType
    user_id: int
    beatmap: Optional[Beatmap] = None
    beatmapset: Optional[Beatmapset] = None
    best_id: Optional[int] = None
    build_id: Optional[int] = None
    current_user_attributes: Optional[CurrentUserAttributes] = None
    id: Optional[int] = None
    match: Optional[MultiplayerMatch] = None
    preserved: Optional[bool] = None
    position: Optional[int] = None
    ranked: Optional[bool] = None
    rank_country: Optional[int] = None
    rank_global: Optional[int] = None
    legacy_score_id: Optional[int] = None
    legacy_total_score: Optional[int] = None
    playlist_item_id: Optional[int] = None
    pp: Optional[float] = None
    room_id: Optional[int] = None
    started_at: Optional[datetime] = None
    user: Optional[User] = None
    weight: Optional[ScoreWeight] = None

    @property
    def created_at(self) -> datetime:
        return self.ended_at

    @property
    def score(self) -> int:
        return self.total_score

    @property
    def replay(self) -> bool:
        """Legacy shortcut for has_replay."""
        return self.has_replay

    @property
    def score_url(self) -> Optional[str]:
        r"""Link to the score.

        :return: Link to the score on the osu! website
        :rtype: Optional[str]
        """
        if not self.id:
            return None
        if self.type == "solo_score":
            return f"https://osu.ppy.sh/scores/{self.id}"

        if not self.best_id:  # Legacy URL format
            return None
        return f"https://osu.ppy.sh/scores/{self.mode.name_api}/{self.best_id}"

    @property
    def replay_url(self) -> Optional[str]:
        r"""Link to the replay.

        :return: Link to download the replay on the osu! website
        :rtype: Optional[str]
        """
        if not self.has_replay:
            return None
        score_url = self.score_url
        if not score_url:
            return None
        return score_url + "/download"

    @computed_field  # type: ignore
    @cached_property
    def completion(self) -> Optional[float]:
        """Beatmap completion.

        :return: Beatmap completion of a score (%). 100% for passes. None if no beatmap.
        :rtype: Optional[float]
        """
        if not self.beatmap:
            return None

        if self.passed:
            return 100.0

        return calculate_score_completion(self.statistics, self.beatmap)

    @computed_field  # type: ignore
    @cached_property
    def mode(self) -> Gamemode:
        return Gamemode(self.ruleset_id)

    @computed_field  # type: ignore
    @cached_property
    def mods_str(self) -> str:
        return "".join(str(mod) for mod in self.mods)

    @model_validator(mode="before")
    @classmethod
    def _fail_rank(cls, values: dict[str, object]) -> dict[str, object]:
        if not values["passed"]:
            values["rank"] = "F"
        return values


from .multiplayer import MultiplayerMatch



"""
This module contains models for the search endpoint.
"""

from __future__ import annotations

from typing import Literal
from typing import Optional
from typing import Union

from .base import BaseModel
from .user import User
from .wiki import WikiPage

__all__ = (
    "SearchMode",
    "SearchResponse",
    "SearchResult",
)

SearchMode = Literal["all", "user", "wiki_page"]


class SearchResult(BaseModel):
    data: list[Union[User, WikiPage]]
    total: int


class SearchResponse(BaseModel):
    users: Optional[SearchResult] = None
    wiki_pages: Optional[SearchResult] = None
